from typing import Literal

import numpy as np
from pepeline import screentone, cvt_color, CVTColor, DotType, halftone, ResizesAlg
import cv2 as cv


def rgb_halftone(
    img: np.ndarray, dot_size: list[int], angle: list[int], dot_type: list[DotType], scale: float | None, ssaa_filter: ResizesAlg
) -> np.ndarray:
    dot_size_len = len(dot_size)
    angle_len = len(angle)
    dot_type_len = len(dot_type)
    if img.ndim == 2:
        img = cvt_color(img, CVTColor.Gray2RGB)
    img = halftone(
        img,
        [dot_size[index % dot_size_len] for index in range(3)],
        [angle[index % angle_len] for index in range(3)],
        [dot_type[index % dot_type_len] for index in range(3)],
        scale,
        ssaa_filter,
    )
    return img


def cmyk_halftone(
    img: np.ndarray, dot_size: list[int], angle: list[int], dot_type: list[DotType], scale: float | None, ssaa_filter: ResizesAlg
) -> np.ndarray:
    dot_size_len = len(dot_size)
    angle_len = len(angle)
    dot_type_len = len(dot_type)
    if img.ndim == 2:
        img = cvt_color(img, CVTColor.Gray2RGB)
    img = cvt_color(img, CVTColor.RGB2CMYK)
    img = halftone(
        img,
        [dot_size[index % dot_size_len] for index in range(4)],
        [angle[index % angle_len] for index in range(4)],
        [dot_type[index % dot_type_len] for index in range(4)],
        scale,
        ssaa_filter,
    )
    img = cvt_color(img, CVTColor.CMYK2RGB)
    return img


def hsv_halftone(
    img: np.ndarray, dot_size: list[int], angle: list[int], dot_type: list[DotType], scale: float | None, ssaa_filter: ResizesAlg
) -> np.ndarray:
    if img.ndim == 2:
        img = cv.cvtColor(img, cv.COLOR_GRAY2RGB)
    img = cv.cvtColor(img, cv.COLOR_RGB2HSV)
    img[..., 2] = screentone(img[..., 2], dot_size[0], angle[0], dot_type[0], scale, ssaa_filter)
    img = cv.cvtColor(img, cv.COLOR_HSV2RGB)
    return img


def gray_halftone(
    img: np.ndarray, dot_size: list[int], angle: list[int], dot_type: list[DotType], scale: float | None, ssaa_filter: ResizesAlg
) -> np.ndarray:
    if img.ndim == 3:
        img = cvt_color(img, CVTColor.RGB2Gray_2020)
    if scale:
        black_mask = 1 - (img == 0).astype(np.float32)
        white_mask = (img == 1).astype(np.float32)
        return (screentone(img, dot_size[0], angle[0], dot_type[0], scale, ssaa_filter) + white_mask) * black_mask
    else:
        return screentone(img, dot_size[0], angle[0], dot_type[0], scale, ssaa_filter)


MODE_MAP = {'rgb': rgb_halftone, 'gray': gray_halftone, 'hsv': hsv_halftone, 'cmyk': cmyk_halftone}

Mode = Literal['rgb', 'gray', 'hsv', 'cmyk']



from __future__ import annotations

import os.path
from dataclasses import dataclass
from typing import List, Literal

import numpy as np
from pepeline import save

from reline.static import Node, NodeOptions, ImageFile

FileFormat = Literal['png', 'jpeg']


@dataclass(frozen=True)
class FolderWriterOptions(NodeOptions):
    path: str
    format: FileFormat = 'png'


class FolderWriterNode(Node[FolderWriterOptions]):
    def __init__(self, options: FolderWriterOptions):
        super().__init__(options)

        os.makedirs(options.path, exist_ok=True)

    def process(self, files: List[ImageFile]):
        for file in files:
            full_path = os.path.join(os.path.abspath(self.options.path), file.dir, f'{file.basename}.{self.options.format}')
            os.makedirs(os.path.dirname(full_path), exist_ok=True)
            save(file.data, full_path)

    def single_process(self, file: ImageFile):
        full_path = os.path.join(os.path.abspath(self.options.path), file.dir, f'{file.basename}.{self.options.format}')
        os.makedirs(os.path.dirname(full_path), exist_ok=True)
        save(file.data, full_path)

    def video_process(self, file: np.ndarray) -> np.ndarray:
        raise ValueError('Video scale does not support folder writer')



from typing import Literal

from pepeline import ResizesAlg, ResizesFilter

FILTER_MAP = {
    'nearest': ResizesAlg.Nearest(),
    'box': ResizesAlg.Conv(ResizesFilter.Box),
    'sbox4': ResizesAlg.SuperSampling(ResizesFilter.Box, 4),
    'sbox8': ResizesAlg.SuperSampling(ResizesFilter.Box, 8),
    'linear': ResizesAlg.Conv(ResizesFilter.Bilinear),
    'slinear4': ResizesAlg.SuperSampling(ResizesFilter.Bilinear, 4),
    'slinear8': ResizesAlg.SuperSampling(ResizesFilter.Bilinear, 8),
    'hamming': ResizesAlg.Conv(ResizesFilter.Hamming),
    'shamming4': ResizesAlg.SuperSampling(ResizesFilter.Hamming, 4),
    'shamming8': ResizesAlg.SuperSampling(ResizesFilter.Hamming, 8),
    'catmullrom': ResizesAlg.Conv(ResizesFilter.CatmullRom),
    'scatmullrom4': ResizesAlg.SuperSampling(ResizesFilter.CatmullRom, 4),
    'scatmullrom8': ResizesAlg.SuperSampling(ResizesFilter.CatmullRom, 8),
    'mitchell': ResizesAlg.Conv(ResizesFilter.Mitchell),
    'smitchell4': ResizesAlg.SuperSampling(ResizesFilter.Mitchell, 4),
    'smitchell8': ResizesAlg.SuperSampling(ResizesFilter.Mitchell, 8),
    'lanczos': ResizesAlg.Conv(ResizesFilter.Lanczos3),
    'slanczos4': ResizesAlg.SuperSampling(ResizesFilter.Lanczos3, 4),
    'slanczos8': ResizesAlg.SuperSampling(ResizesFilter.Lanczos3, 8),
    'gauss': ResizesAlg.Conv(ResizesFilter.Gaussian),
    'sgauss4': ResizesAlg.SuperSampling(ResizesFilter.Gaussian, 4),
    'sgauss8': ResizesAlg.SuperSampling(ResizesFilter.Gaussian, 8),
}

FilterType = Literal[
    'nearest',
    'box',
    'sbox4',
    'sbox8',
    'linear',
    'slinear4',
    'slinear8',
    'hamming',
    'shamming4',
    'shamming8',
    'catmullrom',
    'scatmullrom4',
    'scatmullrom8',
    'mitchell',
    'smitchell4',
    'smitchell8',
    'lanczos',
    'slanczos4',
    'slanczos8',
    'gauss',
    'sgauss4',
    'sgauss8',
]



from dataclasses import dataclass
from typing import List, Optional, Literal

import numpy as np
from pepeline import cvt_color, CVTColor
from ._sharp_class import Canny, DiapasonBlack, DiapasonWhite, ColorLevels
from reline.static import Node, NodeOptions, ImageFile

CannyType = Literal['invert', 'normal', 'unsharp']


@dataclass(frozen=True)
class SharpOptions(NodeOptions):
    low_input: Optional[int] = 0
    high_input: Optional[int] = 255
    gamma: Optional[float] = 1.0
    diapason_white: Optional[int] = -1
    diapason_black: Optional[int] = -1
    canny: Optional[bool] = False
    canny_type: Optional[CannyType] = 'normal'


class SharpNode(Node[SharpOptions]):
    def __init__(self, options):
        super().__init__(options)
        self.stack = []
        if options.low_input != 0 or options.high_input != 255 or options.gamma != 1.0:
            self.stack.append(ColorLevels(options.low_input, options.high_input, options.gamma))
        if options.diapason_white >= 0:
            self.stack.append(DiapasonWhite(options.diapason_white))
        if options.diapason_black >= 0:
            self.stack.append(DiapasonBlack(options.diapason_black))
        if options.canny:
            self.stack.append(Canny(options.canny_type))

    def process(self, files: List[ImageFile]) -> List[ImageFile]:
        if len(self.stack) == 0:
            return files
        for file in files:
            img_float = file.data.squeeze()
            if img_float.ndim == 3:
                img_float = cvt_color(img_float, CVTColor.RGB2Gray_2020)
            for process in self.stack:
                img_float = process.run(img_float)
            file.data = img_float
        return files

    def single_process(self, file: ImageFile) -> ImageFile:
        if len(self.stack) == 0:
            return file

        img_float = file.data.squeeze()
        if img_float.ndim == 3:
            img_float = cvt_color(img_float, CVTColor.RGB2Gray_2020)
        for process in self.stack:
            img_float = process.run(img_float)
        file.data = img_float
        return file

    def video_process(self, file: np.ndarray) -> np.ndarray:
        if len(self.stack) == 0:
            return file

        img_float = file.squeeze()
        if img_float.ndim == 3:
            img_float = cvt_color(img_float, CVTColor.RGB2Gray_2020)
        for process in self.stack:
            img_float = process.run(img_float)
        return img_float



from typing import Literal, List

import numpy as np

from ...static import NodeOptions, Node, ImageFile
from dataclasses import dataclass
from pepeline import CVTColor, cvt_color

CvtDict = {
    'RGB2Gray2020': CVTColor.RGB2Gray_2020,
    'RGB2Gray709': CVTColor.RGB2Gray_709,
    'RGB2Gray': CVTColor.RGB2Gray_601,
    'Gray2RGB': CVTColor.Gray2RGB,
}
CvtTypeList = Literal['RGB2Gray2020', 'RGB2Gray709', 'RGB2Gray', 'Gray2RGB']


@dataclass(frozen=True)
class CvtColorOptions(NodeOptions):
    cvt_type: CvtTypeList


class CvtColorNode(Node[CvtColorOptions]):
    def __init__(self, options):
        super().__init__(options)
        self.cvt_type = CvtDict[options.cvt_type]

    def __cvt_logic(self, img: np.ndarray) -> np.ndarray:
        if self.cvt_type in [CVTColor.RGB2Gray_2020, CVTColor.RGB2Gray_709, CVTColor.RGB2Gray_601] and img.ndim == 3:
            return cvt_color(img, self.cvt_type)
        elif self.cvt_type in [CVTColor.Gray2RGB] and img.ndim == 2:
            return cvt_color(img, self.cvt_type)
        else:
            return img

    def process(self, files: List[ImageFile]) -> List[ImageFile]:
        for file in files:
            file.data = self.__cvt_logic(file.data)

        return files

    def single_process(self, file: ImageFile) -> ImageFile:
        file.data = self.__cvt_logic(file.data)

        return file

    def video_process(self, file: np.ndarray) -> np.ndarray:
        file = self.__cvt_logic(file)

        return file



import cv2 as cv
import numpy as np

from pepeline import color_levels


class Canny:
    def __init__(self, canny_type):
        self.canny_type = canny_type

    def run(self, img_float: np.ndarray) -> np.ndarray:
        image = (img_float * 255).astype(np.uint8)
        edges = np.clip(255 - cv.Canny(image, 750, 800, apertureSize=3, L2gradient=False), 0, 1)
        if self.canny_type == 'unsharp':
            kernel = np.ones((5, 5), dtype=np.uint8)
            edges = cv.dilate(edges, kernel, iterations=1)
            blurred = cv.GaussianBlur(img_float, (0, 0), sigmaX=0.5, sigmaY=0.5, borderType=cv.BORDER_REFLECT)
            white = ((img_float * edges) > 0.9882352941176471).astype(np.float32)
            black = (img_float < 0.011764705882352941).astype(np.float32) * edges
            black = cv.dilate(black, kernel, iterations=1)
            white = cv.dilate(white, kernel, iterations=1)
            w_b = ((white + black) - 1).clip(0, 1)
            img_float = (cv.addWeighted(img_float, 11, blurred, -10, 0) * w_b + img_float * (1 - w_b)).clip(0, 1)

            return img_float
        else:
            return np.where(edges, img_float, 1.0 if self.canny_type == 'invert' else 0.0)


class DiapasonWhite:
    def __init__(self, diapason_white: int):
        self.diapason = diapason_white / 255

    def run(self, img_float: np.ndarray) -> np.ndarray:
        image = (img_float * 255).astype(np.uint8)
        median_image = cv.medianBlur(image, 3)
        mask = median_image <= 255 - self.diapason
        return np.where(mask, img_float, 1.0)


class DiapasonBlack:
    def __init__(self, diapason_black: int):
        self.diapason = diapason_black / 255

    def run(self, img_float: np.ndarray) -> np.ndarray:
        black_mask = (img_float > self.diapason).astype(np.float32)
        blur = cv.GaussianBlur(black_mask, (3, 3), 0)
        blur_mask = (blur > 0.9).astype(np.float32)
        return np.where(blur_mask, img_float, 0)


class ColorLevels:
    def __init__(self, low_input, high_input, gamma):
        self.low_input = low_input
        self.high_input = high_input
        self.gamma = gamma

    def run(self, img_float: np.ndarray) -> np.ndarray:
        return color_levels(img_float, self.low_input, self.high_input, 0, 255, self.gamma)
